# DragonLog © 2023-2024 by Andreas Schawo is licensed under CC BY-SA 4.0.
# To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/4.0/

import re
import csv
from collections import namedtuple
from collections.abc import Generator

Country = namedtuple('Country', ('code', 'name', 'dxcc',
                                 'continent', 'cq', 'itu',
                                 'lat', 'lon', 'time_off'))


class CountryCodeNotFoundException(Exception):
    pass


class CountryNotFoundException(Exception):
    pass


class CountryData:
    """Loads country data from a cty CSV file
    Provides search for prefix, country code and country data by a given callsign"""

    # noinspection RegExpRedundantEscape
    RE_OVERRIDES = re.compile(r'([A-Z0-9/=]+)(?:\((\d+)\))?(?:\[(\d+)\])?.*')

    def __init__(self, cty_file):
        self.__countries__: dict[str, Country] = {}
        self.__prefixes__ = {}
        self.__pfx_list__ = []
        self.__calls__ = {}
        self.__data_ver__ = ''

        self.__load__(cty_file)

    def __load__(self, cty_file: str):
        """Loads the cty data file and builds the internal data structure
        :param cty_file: the cty data in CSV format"""

        with (open(cty_file) as cty_f):
            cty = csv.reader(cty_f, delimiter=',')
            for row in cty:
                row[0] = row[0].replace('*', '')
                self.__countries__[row[0]] = Country(*row[:-1])

                for pfx in row[9][:-1].split():
                    data = {'cty_code': row[0]}

                    pfx_parts = re.findall(self.RE_OVERRIDES, pfx)[0]
                    pfx, cq_or, itu_or = pfx_parts
                    if cq_or:
                        data['cq_or'] = cq_or[0]
                    if itu_or:
                        data['itu_or'] = itu_or[0]

                    if pfx.startswith('='):
                        if row[0] == 'VE' and pfx.startswith('=VER'):
                            self.__data_ver__ = pfx[1:]
                            continue
                        self.__calls__[pfx[1:]] = data
                    else:
                        self.__prefixes__[pfx] = data

        self.__pfx_list__ = sorted(self.__prefixes__)

    @property
    def version(self) -> str:
        """Get the version of the data file"""

        return self.__data_ver__

    def prefix(self, call: str) -> str:
        """Get the prefix of a callsign
        :param call: the callsign (upper or lowercase)
        :return: the prefix"""

        part = ''
        i = 0
        pfx = ''
        found = False
        for d in call.upper():
            part += d
            try:
                i = self.__pfx_list__.index(part, i)
                pfx = self.__pfx_list__[i]
            except ValueError:
                if found:
                    break
                if pfx:
                    found = True
                continue

        return pfx

    def cty_code(self, call) -> dict:
        """Get the country code for a callsign
        :param call: the callsign (upper or lowercase)
        :return: the country code"""

        if call in self.__calls__:
            return self.__calls__[call.upper()]
        elif self.prefix(call.upper()):
            return self.__prefixes__[self.prefix(call.upper())]
        else:
            raise CountryCodeNotFoundException(f'for "{call}"')

    def country(self, call: str) -> Country:
        """Get the country data from a callsign
        :param call: the callsign (upper or lowercase)
        :return: the country data"""

        data = self.cty_code(call)
        if data:
            cty_code = data['cty_code']
            cty_data = self.__countries__[cty_code]
            if 'cq_or' in data:
                cty_data = cty_data._replace(cq=data['cq_or'])
            if 'itu_or' in data:
                cty_data = cty_data._replace(itu=data['itu_or'])

            return cty_data
        else:
            raise CountryNotFoundException(f'for "{call}"')

    @property
    def countries(self) -> Generator:
        for c in self.__countries__.values():
            yield c.name


def main():
    cty = CountryData('data/cty/cty.csv')
    print(cty.version)
    print(cty.country('VERSION'))


if __name__ == '__main__':
    main()



import os

from PyQt6 import QtWidgets, QtCore

from . import DragonLog_AppSelect_ui

class AppSelect(QtWidgets.QDialog, DragonLog_AppSelect_ui.Ui_appSelectDialog):
    def __init__(self, parent, title: str, settings: QtCore.QSettings):
        super().__init__(parent)
        self.setupUi(self)
        self.setWindowTitle(title)
        self.settings = settings

    def exec(self) -> tuple[str, str]:
        if super().exec() == 1:
            if self.wsjtxRadioButton.isChecked():
                if os.name == 'posix':
                    path = os.path.abspath(os.path.expanduser('~/.local/share/WSJT-X/wsjtx_log.adi'))
                elif os.name == 'nt':
                    path = os.path.abspath(os.path.expanduser('~/AppData/Local/WSJT-X/wsjtx_log.adi'))
                else:
                    path = os.path.abspath(os.curdir)
                return 'WSJTX', self.settings.value('lastWatchFileWSJTX', path)
            elif self.js8callRadioButton.isChecked():
                if os.name == 'posix':
                    path = os.path.abspath(os.path.expanduser('~/.local/share/JS8Call/js8call_log.adi'))
                elif os.name == 'nt':
                    path = os.path.abspath(os.path.expanduser('~/AppData/Local/JS8Call/js8call_log.adi'))
                else:
                    path = os.path.abspath(os.curdir)
                return 'JS8Call', self.settings.value('lastWatchFileJS8Call', path)
            elif self.fldigiRadioButton.isChecked():
                if os.name == 'posix':
                    path = os.path.abspath(os.path.expanduser('~/.fldigi/logs/logbook.adi'))
                elif os.name == 'nt':
                    path = os.path.abspath(os.path.expanduser('~/fldigi.files/logs/logbook.adi'))
                else:
                    path = os.path.abspath(os.curdir)
                return 'fldigi', self.settings.value('lastWatchFilefldigi', path)
            else:
                return 'Other', self.settings.value('lastWatchFile', os.path.abspath(os.curdir))



"""Module contains the regular expression for field validation"""
import re

REGEX_CALL = re.compile(r'([a-zA-Z0-9]{1,3}?/)?([a-zA-Z0-9]{1,3}?[0-9][a-zA-Z0-9]{0,3}?[a-zA-Z])(/[aAmMpPrRtT]{1,2}?)?')
REGEX_RSTFIELD = re.compile(r'([1-5]([1-9]([1-9][aAcCkKmMsSxX]?)?)?)|([-+][0-9]{1,2})')
REGEX_LOCATOR = re.compile(r'[a-rA-R]{2}[0-9]{2}([a-xA-X]{2}([0-9]{2})?)?')
REGEX_NONASCII = re.compile(r'[ -~\n\r]*(.)?')
REGEX_TIME = re.compile(r'(([0-1][0-9])|(2[0-3])):([0-5][0-9])(:[0-5][0-9])?')
REGEX_DATE = re.compile(r'([1-9][0-9]{3})-((0[1-9])|(1[0-2]))-((0[1-9])|([1-2][0-9])|(3[0-1]))')
REGEX_QTH = re.compile(r'(.*?)? *\(([a-rA-R]{2}[0-9]{2}([a-xA-X]{2}([0-9]{2})?)?)\)')
# noinspection RegExpRedundantEscape
REGEX_EMAIL = re.compile(r'[\w\-\.]+@([\w-]+\.)+[\w-]{2,}') # r'.+@.+\..+')

def check_format(exp: re.Pattern, txt: str) -> bool:
    """Test the given text against a regular expression
    :param exp: a compiled pattern
    :param txt: a text
    :return: true if pattern matches"""
    return bool(re.fullmatch(exp, txt))


def check_call(call: str) -> None | tuple:
    """Test a call sign against a regular expression
    :param call: a call sign
    :return: tuple of parts ('Country prefix/', 'Call sign', '/Operation suffix')"""

    m = re.fullmatch(REGEX_CALL, call)
    if m:
        return m.groups()


def find_non_ascii(text: str) -> set:
    non_ascii = []
    for c in re.findall(REGEX_NONASCII, text):
        if c:
            non_ascii.append(c)
    return set(non_ascii)


def check_qth(qth_loc: str) -> None | tuple:
    """Test a QTH + locator against a regular expression
    :param qth_loc: "QTH (locator)"
    :return: tuple of parts ('QTH', 'locator')"""

    m = re.fullmatch(REGEX_QTH, qth_loc.strip())
    if m:
        return m.groups()[:2]



# Form implementation generated from reading ui file 'DragonLog_MainWindow.ui'
#
# Created by: PyQt6 UI code generator 6.9.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1286, 900)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("icons:icons8-dragon-96.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        MainWindow.setWindowIcon(icon)
        self.centralwidget = QtWidgets.QWidget(parent=MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.QSOTableView = QtWidgets.QTableView(parent=self.centralwidget)
        self.QSOTableView.setEditTriggers(QtWidgets.QAbstractItemView.EditTrigger.NoEditTriggers)
        self.QSOTableView.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)
        self.QSOTableView.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
        self.QSOTableView.setSortingEnabled(True)
        self.QSOTableView.setObjectName("QSOTableView")
        self.QSOTableView.verticalHeader().setVisible(False)
        self.verticalLayout.addWidget(self.QSOTableView)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(parent=MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1286, 22))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(parent=self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuEdit = QtWidgets.QMenu(parent=self.menubar)
        self.menuEdit.setObjectName("menuEdit")
        self.menuHelp = QtWidgets.QMenu(parent=self.menubar)
        self.menuHelp.setObjectName("menuHelp")
        self.menuUpload = QtWidgets.QMenu(parent=self.menubar)
        self.menuUpload.setObjectName("menuUpload")
        self.menuContest = QtWidgets.QMenu(parent=self.menubar)
        self.menuContest.setObjectName("menuContest")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(parent=MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.toolBar = QtWidgets.QToolBar(parent=MainWindow)
        self.toolBar.setObjectName("toolBar")
        MainWindow.addToolBar(QtCore.Qt.ToolBarArea.TopToolBarArea, self.toolBar)
        self.logDockWidget = QtWidgets.QDockWidget(parent=MainWindow)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.logDockWidget.sizePolicy().hasHeightForWidth())
        self.logDockWidget.setSizePolicy(sizePolicy)
        self.logDockWidget.setFeatures(QtWidgets.QDockWidget.DockWidgetFeature.DockWidgetClosable|QtWidgets.QDockWidget.DockWidgetFeature.DockWidgetFloatable|QtWidgets.QDockWidget.DockWidgetFeature.DockWidgetMovable)
        self.logDockWidget.setAllowedAreas(QtCore.Qt.DockWidgetArea.AllDockWidgetAreas)
        self.logDockWidget.setObjectName("logDockWidget")
        self.logDockWidgetContents = QtWidgets.QWidget()
        self.logDockWidgetContents.setObjectName("logDockWidgetContents")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.logDockWidgetContents)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.logTextEdit = QtWidgets.QTextEdit(parent=self.logDockWidgetContents)
        font = QtGui.QFont()
        font.setFamily("Consolas")
        self.logTextEdit.setFont(font)
        self.logTextEdit.setStyleSheet("font-family: Consolas")
        self.logTextEdit.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.TextSelectableByKeyboard|QtCore.Qt.TextInteractionFlag.TextSelectableByMouse)
        self.logTextEdit.setObjectName("logTextEdit")
        self.verticalLayout_2.addWidget(self.logTextEdit)
        self.logDockWidget.setWidget(self.logDockWidgetContents)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(8), self.logDockWidget)
        self.qsoDockWidget = QtWidgets.QDockWidget(parent=MainWindow)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.qsoDockWidget.sizePolicy().hasHeightForWidth())
        self.qsoDockWidget.setSizePolicy(sizePolicy)
        self.qsoDockWidget.setAllowedAreas(QtCore.Qt.DockWidgetArea.LeftDockWidgetArea|QtCore.Qt.DockWidgetArea.RightDockWidgetArea)
        self.qsoDockWidget.setObjectName("qsoDockWidget")
        self.qsoDockWidgetContents = QtWidgets.QWidget()
        self.qsoDockWidgetContents.setObjectName("qsoDockWidgetContents")
        self.qsoDockWidget.setWidget(self.qsoDockWidgetContents)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(2), self.qsoDockWidget)
        self.filterDockWidget = QtWidgets.QDockWidget(parent=MainWindow)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.filterDockWidget.sizePolicy().hasHeightForWidth())
        self.filterDockWidget.setSizePolicy(sizePolicy)
        self.filterDockWidget.setFeatures(QtWidgets.QDockWidget.DockWidgetFeature.DockWidgetClosable|QtWidgets.QDockWidget.DockWidgetFeature.DockWidgetFloatable|QtWidgets.QDockWidget.DockWidgetFeature.DockWidgetMovable)
        self.filterDockWidget.setAllowedAreas(QtCore.Qt.DockWidgetArea.BottomDockWidgetArea|QtCore.Qt.DockWidgetArea.TopDockWidgetArea)
        self.filterDockWidget.setObjectName("filterDockWidget")
        self.filterDockWidgetContents = QtWidgets.QWidget()
        self.filterDockWidgetContents.setObjectName("filterDockWidgetContents")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout(self.filterDockWidgetContents)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout()
        self.verticalLayout_6.setSpacing(0)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.label_4 = QtWidgets.QLabel(parent=self.filterDockWidgetContents)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_6.addWidget(self.label_4)
        self.fDateFromLineEdit = QtWidgets.QLineEdit(parent=self.filterDockWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fDateFromLineEdit.sizePolicy().hasHeightForWidth())
        self.fDateFromLineEdit.setSizePolicy(sizePolicy)
        self.fDateFromLineEdit.setMaximumSize(QtCore.QSize(90, 16777215))
        self.fDateFromLineEdit.setObjectName("fDateFromLineEdit")
        self.verticalLayout_6.addWidget(self.fDateFromLineEdit)
        self.horizontalLayout_2.addLayout(self.verticalLayout_6)
        self.verticalLayout_9 = QtWidgets.QVBoxLayout()
        self.verticalLayout_9.setSpacing(0)
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.label_5 = QtWidgets.QLabel(parent=self.filterDockWidgetContents)
        self.label_5.setObjectName("label_5")
        self.verticalLayout_9.addWidget(self.label_5)
        self.fDateToLineEdit = QtWidgets.QLineEdit(parent=self.filterDockWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fDateToLineEdit.sizePolicy().hasHeightForWidth())
        self.fDateToLineEdit.setSizePolicy(sizePolicy)
        self.fDateToLineEdit.setMaximumSize(QtCore.QSize(90, 16777215))
        self.fDateToLineEdit.setObjectName("fDateToLineEdit")
        self.verticalLayout_9.addWidget(self.fDateToLineEdit)
        self.horizontalLayout_2.addLayout(self.verticalLayout_9)
        spacerItem = QtWidgets.QSpacerItem(10, 20, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label = QtWidgets.QLabel(parent=self.filterDockWidgetContents)
        self.label.setObjectName("label")
        self.verticalLayout_3.addWidget(self.label)
        self.fCallsignLineEdit = QtWidgets.QLineEdit(parent=self.filterDockWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fCallsignLineEdit.sizePolicy().hasHeightForWidth())
        self.fCallsignLineEdit.setSizePolicy(sizePolicy)
        self.fCallsignLineEdit.setObjectName("fCallsignLineEdit")
        self.verticalLayout_3.addWidget(self.fCallsignLineEdit)
        self.horizontalLayout_2.addLayout(self.verticalLayout_3)
        spacerItem1 = QtWidgets.QSpacerItem(10, 20, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.verticalLayout_5 = QtWidgets.QVBoxLayout()
        self.verticalLayout_5.setSpacing(0)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.label_3 = QtWidgets.QLabel(parent=self.filterDockWidgetContents)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_5.addWidget(self.label_3)
        self.fBandComboBox = QtWidgets.QComboBox(parent=self.filterDockWidgetContents)
        self.fBandComboBox.setObjectName("fBandComboBox")
        self.verticalLayout_5.addWidget(self.fBandComboBox)
        self.horizontalLayout_2.addLayout(self.verticalLayout_5)
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setSpacing(0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label_2 = QtWidgets.QLabel(parent=self.filterDockWidgetContents)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_4.addWidget(self.label_2)
        self.fModeComboBox = QtWidgets.QComboBox(parent=self.filterDockWidgetContents)
        self.fModeComboBox.setObjectName("fModeComboBox")
        self.verticalLayout_4.addWidget(self.fModeComboBox)
        self.horizontalLayout_2.addLayout(self.verticalLayout_4)
        spacerItem2 = QtWidgets.QSpacerItem(10, 20, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem2)
        self.verticalLayout_7 = QtWidgets.QVBoxLayout()
        self.verticalLayout_7.setSpacing(0)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.label_6 = QtWidgets.QLabel(parent=self.filterDockWidgetContents)
        self.label_6.setObjectName("label_6")
        self.verticalLayout_7.addWidget(self.label_6)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.fQSLsComboBox = QtWidgets.QComboBox(parent=self.filterDockWidgetContents)
        self.fQSLsComboBox.setObjectName("fQSLsComboBox")
        self.fQSLsComboBox.addItem("")
        self.fQSLsComboBox.setItemText(0, "")
        self.fQSLsComboBox.addItem("")
        self.fQSLsComboBox.addItem("")
        self.horizontalLayout.addWidget(self.fQSLsComboBox)
        self.fQSLrComboBox = QtWidgets.QComboBox(parent=self.filterDockWidgetContents)
        self.fQSLrComboBox.setObjectName("fQSLrComboBox")
        self.fQSLrComboBox.addItem("")
        self.fQSLrComboBox.setItemText(0, "")
        self.fQSLrComboBox.addItem("")
        self.fQSLrComboBox.addItem("")
        self.horizontalLayout.addWidget(self.fQSLrComboBox)
        self.verticalLayout_7.addLayout(self.horizontalLayout)
        self.horizontalLayout_2.addLayout(self.verticalLayout_7)
        self.verticalLayout_10 = QtWidgets.QVBoxLayout()
        self.verticalLayout_10.setSpacing(0)
        self.verticalLayout_10.setObjectName("verticalLayout_10")
        self.label_7 = QtWidgets.QLabel(parent=self.filterDockWidgetContents)
        self.label_7.setObjectName("label_7")
        self.verticalLayout_10.addWidget(self.label_7)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.feQSLsComboBox = QtWidgets.QComboBox(parent=self.filterDockWidgetContents)
        self.feQSLsComboBox.setObjectName("feQSLsComboBox")
        self.feQSLsComboBox.addItem("")
        self.feQSLsComboBox.setItemText(0, "")
        self.feQSLsComboBox.addItem("")
        self.feQSLsComboBox.addItem("")
        self.horizontalLayout_3.addWidget(self.feQSLsComboBox)
        self.feQSLrComboBox = QtWidgets.QComboBox(parent=self.filterDockWidgetContents)
        self.feQSLrComboBox.setObjectName("feQSLrComboBox")
        self.feQSLrComboBox.addItem("")
        self.feQSLrComboBox.setItemText(0, "")
        self.feQSLrComboBox.addItem("")
        self.feQSLrComboBox.addItem("")
        self.horizontalLayout_3.addWidget(self.feQSLrComboBox)
        self.verticalLayout_10.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_2.addLayout(self.verticalLayout_10)
        self.verticalLayout_11 = QtWidgets.QVBoxLayout()
        self.verticalLayout_11.setSpacing(0)
        self.verticalLayout_11.setObjectName("verticalLayout_11")
        self.label_8 = QtWidgets.QLabel(parent=self.filterDockWidgetContents)
        self.label_8.setObjectName("label_8")
        self.verticalLayout_11.addWidget(self.label_8)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.fLoTWsComboBox = QtWidgets.QComboBox(parent=self.filterDockWidgetContents)
        self.fLoTWsComboBox.setObjectName("fLoTWsComboBox")
        self.fLoTWsComboBox.addItem("")
        self.fLoTWsComboBox.setItemText(0, "")
        self.fLoTWsComboBox.addItem("")
        self.fLoTWsComboBox.addItem("")
        self.horizontalLayout_4.addWidget(self.fLoTWsComboBox)
        self.fLoTWrComboBox = QtWidgets.QComboBox(parent=self.filterDockWidgetContents)
        self.fLoTWrComboBox.setObjectName("fLoTWrComboBox")
        self.fLoTWrComboBox.addItem("")
        self.fLoTWrComboBox.setItemText(0, "")
        self.fLoTWrComboBox.addItem("")
        self.fLoTWrComboBox.addItem("")
        self.horizontalLayout_4.addWidget(self.fLoTWrComboBox)
        self.verticalLayout_11.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_2.addLayout(self.verticalLayout_11)
        spacerItem3 = QtWidgets.QSpacerItem(10, 20, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem3)
        self.verticalLayout_12 = QtWidgets.QVBoxLayout()
        self.verticalLayout_12.setSpacing(0)
        self.verticalLayout_12.setObjectName("verticalLayout_12")
        self.label_9 = QtWidgets.QLabel(parent=self.filterDockWidgetContents)
        self.label_9.setObjectName("label_9")
        self.verticalLayout_12.addWidget(self.label_9)
        self.fContestComboBox = QtWidgets.QComboBox(parent=self.filterDockWidgetContents)
        self.fContestComboBox.setEditable(True)
        self.fContestComboBox.setObjectName("fContestComboBox")
        self.verticalLayout_12.addWidget(self.fContestComboBox)
        self.horizontalLayout_2.addLayout(self.verticalLayout_12)
        spacerItem4 = QtWidgets.QSpacerItem(10, 20, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem4)
        self.resetFilterPushButton = QtWidgets.QPushButton(parent=self.filterDockWidgetContents)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap("icons:no.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.resetFilterPushButton.setIcon(icon1)
        self.resetFilterPushButton.setObjectName("resetFilterPushButton")
        self.horizontalLayout_2.addWidget(self.resetFilterPushButton)
        spacerItem5 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem5)
        self.verticalLayout_8.addLayout(self.horizontalLayout_2)
        spacerItem6 = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout_8.addItem(spacerItem6)
        self.filterDockWidget.setWidget(self.filterDockWidgetContents)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(4), self.filterDockWidget)
        self.ccDockWidget = QtWidgets.QDockWidget(parent=MainWindow)
        self.ccDockWidget.setAllowedAreas(QtCore.Qt.DockWidgetArea.BottomDockWidgetArea|QtCore.Qt.DockWidgetArea.TopDockWidgetArea)
        self.ccDockWidget.setObjectName("ccDockWidget")
        self.ccDockWidgetContents = QtWidgets.QWidget()
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ccDockWidgetContents.sizePolicy().hasHeightForWidth())
        self.ccDockWidgetContents.setSizePolicy(sizePolicy)
        self.ccDockWidgetContents.setObjectName("ccDockWidgetContents")
        self.ccDockWidget.setWidget(self.ccDockWidgetContents)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(8), self.ccDockWidget)
        self.dxSpotsDockWidget = QtWidgets.QDockWidget(parent=MainWindow)
        self.dxSpotsDockWidget.setAllowedAreas(QtCore.Qt.DockWidgetArea.LeftDockWidgetArea|QtCore.Qt.DockWidgetArea.RightDockWidgetArea)
        self.dxSpotsDockWidget.setObjectName("dxSpotsDockWidget")
        self.dockWidgetContents = QtWidgets.QWidget()
        self.dockWidgetContents.setObjectName("dockWidgetContents")
        self.dxSpotsDockWidget.setWidget(self.dockWidgetContents)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(1), self.dxSpotsDockWidget)
        self.contestStatDockWidget = QtWidgets.QDockWidget(parent=MainWindow)
        self.contestStatDockWidget.setObjectName("contestStatDockWidget")
        self.dockWidgetContents_2 = QtWidgets.QWidget()
        self.dockWidgetContents_2.setObjectName("dockWidgetContents_2")
        self.contestStatDockWidget.setWidget(self.dockWidgetContents_2)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(8), self.contestStatDockWidget)
        self.actionLog_QSO = QtGui.QAction(parent=MainWindow)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap("icons:edit_add.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionLog_QSO.setIcon(icon2)
        self.actionLog_QSO.setObjectName("actionLog_QSO")
        self.actionLog_QSO_TB = QtGui.QAction(parent=MainWindow)
        self.actionLog_QSO_TB.setIcon(icon2)
        self.actionLog_QSO_TB.setObjectName("actionLog_QSO_TB")
        self.actionSelect_DB = QtGui.QAction(parent=MainWindow)
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap("icons:db.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionSelect_DB.setIcon(icon3)
        self.actionSelect_DB.setObjectName("actionSelect_DB")
        self.actionSelect_DB_TB = QtGui.QAction(parent=MainWindow)
        self.actionSelect_DB_TB.setIcon(icon3)
        self.actionSelect_DB_TB.setObjectName("actionSelect_DB_TB")
        self.actionExit = QtGui.QAction(parent=MainWindow)
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap("icons:exit.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionExit.setIcon(icon4)
        self.actionExit.setObjectName("actionExit")
        self.actionExport = QtGui.QAction(parent=MainWindow)
        self.actionExport.setEnabled(False)
        icon5 = QtGui.QIcon()
        icon5.addPixmap(QtGui.QPixmap("icons:fileexport.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionExport.setIcon(icon5)
        self.actionExport.setObjectName("actionExport")
        self.actionImport = QtGui.QAction(parent=MainWindow)
        icon6 = QtGui.QIcon()
        icon6.addPixmap(QtGui.QPixmap("icons:fileimport.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionImport.setIcon(icon6)
        self.actionImport.setObjectName("actionImport")
        self.actionExport_TB = QtGui.QAction(parent=MainWindow)
        self.actionExport_TB.setEnabled(False)
        self.actionExport_TB.setIcon(icon5)
        self.actionExport_TB.setObjectName("actionExport_TB")
        self.actionImport_TB = QtGui.QAction(parent=MainWindow)
        self.actionImport_TB.setIcon(icon6)
        self.actionImport_TB.setObjectName("actionImport_TB")
        self.actionAbout = QtGui.QAction(parent=MainWindow)
        icon7 = QtGui.QIcon()
        icon7.addPixmap(QtGui.QPixmap("icons:info.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionAbout.setIcon(icon7)
        self.actionAbout.setObjectName("actionAbout")
        self.actionAbout_Qt = QtGui.QAction(parent=MainWindow)
        self.actionAbout_Qt.setIcon(icon7)
        self.actionAbout_Qt.setObjectName("actionAbout_Qt")
        self.actionDelete_log_entry = QtGui.QAction(parent=MainWindow)
        self.actionDelete_log_entry.setEnabled(False)
        icon8 = QtGui.QIcon()
        icon8.addPixmap(QtGui.QPixmap("icons:edit_remove.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionDelete_log_entry.setIcon(icon8)
        self.actionDelete_log_entry.setObjectName("actionDelete_log_entry")
        self.actionChange_log_entry = QtGui.QAction(parent=MainWindow)
        self.actionChange_log_entry.setEnabled(False)
        icon9 = QtGui.QIcon()
        icon9.addPixmap(QtGui.QPixmap("icons:edit.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionChange_log_entry.setIcon(icon9)
        self.actionChange_log_entry.setObjectName("actionChange_log_entry")
        self.actionSettings = QtGui.QAction(parent=MainWindow)
        icon10 = QtGui.QIcon()
        icon10.addPixmap(QtGui.QPixmap("icons:gear.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionSettings.setIcon(icon10)
        self.actionSettings.setObjectName("actionSettings")
        self.actionSettings_TB = QtGui.QAction(parent=MainWindow)
        self.actionSettings_TB.setIcon(icon10)
        self.actionSettings_TB.setObjectName("actionSettings_TB")
        self.actionHelp = QtGui.QAction(parent=MainWindow)
        icon11 = QtGui.QIcon()
        icon11.addPixmap(QtGui.QPixmap("icons:help.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionHelp.setIcon(icon11)
        self.actionHelp.setObjectName("actionHelp")
        self.actionStart_hamlib_TB = QtGui.QAction(parent=MainWindow)
        self.actionStart_hamlib_TB.setCheckable(True)
        icon12 = QtGui.QIcon()
        icon12.addPixmap(QtGui.QPixmap("icons:player_play.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        icon12.addPixmap(QtGui.QPixmap("icons:player_stop.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.On)
        self.actionStart_hamlib_TB.setIcon(icon12)
        self.actionStart_hamlib_TB.setObjectName("actionStart_hamlib_TB")
        self.actionWatch_file_for_QSOs = QtGui.QAction(parent=MainWindow)
        self.actionWatch_file_for_QSOs.setCheckable(True)
        icon13 = QtGui.QIcon()
        icon13.addPixmap(QtGui.QPixmap("icons:watch.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionWatch_file_for_QSOs.setIcon(icon13)
        self.actionWatch_file_for_QSOs.setObjectName("actionWatch_file_for_QSOs")
        self.actionWatch_file_for_QSOs_TB = QtGui.QAction(parent=MainWindow)
        self.actionWatch_file_for_QSOs_TB.setCheckable(True)
        self.actionWatch_file_for_QSOs_TB.setIcon(icon13)
        self.actionWatch_file_for_QSOs_TB.setObjectName("actionWatch_file_for_QSOs_TB")
        self.actionShow_log = QtGui.QAction(parent=MainWindow)
        icon14 = QtGui.QIcon()
        icon14.addPixmap(QtGui.QPixmap("icons:file_doc.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionShow_log.setIcon(icon14)
        self.actionShow_log.setObjectName("actionShow_log")
        self.actionUpload_logs_to_LoTW = QtGui.QAction(parent=MainWindow)
        self.actionUpload_logs_to_LoTW.setEnabled(False)
        icon15 = QtGui.QIcon()
        icon15.addPixmap(QtGui.QPixmap("icons:upload_lotw.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionUpload_logs_to_LoTW.setIcon(icon15)
        self.actionUpload_logs_to_LoTW.setObjectName("actionUpload_logs_to_LoTW")
        self.actionShow_filter = QtGui.QAction(parent=MainWindow)
        icon16 = QtGui.QIcon()
        icon16.addPixmap(QtGui.QPixmap("icons:filter.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionShow_filter.setIcon(icon16)
        self.actionShow_filter.setObjectName("actionShow_filter")
        self.actionShow_CassiopeiaConsole = QtGui.QAction(parent=MainWindow)
        icon17 = QtGui.QIcon()
        icon17.addPixmap(QtGui.QPixmap("icons:terminal.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionShow_CassiopeiaConsole.setIcon(icon17)
        self.actionShow_CassiopeiaConsole.setObjectName("actionShow_CassiopeiaConsole")
        self.actionUpload_to_HamQTH = QtGui.QAction(parent=MainWindow)
        self.actionUpload_to_HamQTH.setEnabled(False)
        icon18 = QtGui.QIcon()
        icon18.addPixmap(QtGui.QPixmap("icons:hamqth.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionUpload_to_HamQTH.setIcon(icon18)
        self.actionUpload_to_HamQTH.setObjectName("actionUpload_to_HamQTH")
        self.actionCheck_LoTW_QSL = QtGui.QAction(parent=MainWindow)
        self.actionCheck_LoTW_QSL.setEnabled(False)
        icon19 = QtGui.QIcon()
        icon19.addPixmap(QtGui.QPixmap("icons:check_lotw.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionCheck_LoTW_QSL.setIcon(icon19)
        self.actionCheck_LoTW_QSL.setObjectName("actionCheck_LoTW_QSL")
        self.actionUpload_to_eQSL = QtGui.QAction(parent=MainWindow)
        self.actionUpload_to_eQSL.setEnabled(False)
        icon20 = QtGui.QIcon()
        icon20.addPixmap(QtGui.QPixmap("icons:upload_eqsl.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionUpload_to_eQSL.setIcon(icon20)
        self.actionUpload_to_eQSL.setObjectName("actionUpload_to_eQSL")
        self.actionCheck_eQSL = QtGui.QAction(parent=MainWindow)
        self.actionCheck_eQSL.setEnabled(False)
        icon21 = QtGui.QIcon()
        icon21.addPixmap(QtGui.QPixmap("icons:check_eqsl.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionCheck_eQSL.setIcon(icon21)
        self.actionCheck_eQSL.setObjectName("actionCheck_eQSL")
        self.actionDownload_eQSL = QtGui.QAction(parent=MainWindow)
        self.actionDownload_eQSL.setEnabled(False)
        icon22 = QtGui.QIcon()
        icon22.addPixmap(QtGui.QPixmap("icons:download_eqsl.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionDownload_eQSL.setIcon(icon22)
        self.actionDownload_eQSL.setObjectName("actionDownload_eQSL")
        self.actionShortcuts = QtGui.QAction(parent=MainWindow)
        icon23 = QtGui.QIcon()
        icon23.addPixmap(QtGui.QPixmap("icons:shortcuts.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionShortcuts.setIcon(icon23)
        self.actionShortcuts.setObjectName("actionShortcuts")
        self.actionHelpCassiopeiaConsole = QtGui.QAction(parent=MainWindow)
        icon24 = QtGui.QIcon()
        icon24.addPixmap(QtGui.QPixmap("icons:help_terminal.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionHelpCassiopeiaConsole.setIcon(icon24)
        self.actionHelpCassiopeiaConsole.setObjectName("actionHelpCassiopeiaConsole")
        self.actionShow_DX_Spots = QtGui.QAction(parent=MainWindow)
        icon25 = QtGui.QIcon()
        icon25.addPixmap(QtGui.QPixmap("icons:find.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionShow_DX_Spots.setIcon(icon25)
        self.actionShow_DX_Spots.setObjectName("actionShow_DX_Spots")
        self.actionExport_Contest_Log = QtGui.QAction(parent=MainWindow)
        icon26 = QtGui.QIcon()
        icon26.addPixmap(QtGui.QPixmap("icons:contest.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionExport_Contest_Log.setIcon(icon26)
        self.actionExport_Contest_Log.setObjectName("actionExport_Contest_Log")
        self.actionContests = QtGui.QAction(parent=MainWindow)
        icon27 = QtGui.QIcon()
        icon27.addPixmap(QtGui.QPixmap("icons:contest_help.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionContests.setIcon(icon27)
        self.actionContests.setObjectName("actionContests")
        self.actionMark_QSL_sent = QtGui.QAction(parent=MainWindow)
        icon28 = QtGui.QIcon()
        icon28.addPixmap(QtGui.QPixmap("icons:ok.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionMark_QSL_sent.setIcon(icon28)
        self.actionMark_QSL_sent.setObjectName("actionMark_QSL_sent")
        self.actionMark_QSL_received = QtGui.QAction(parent=MainWindow)
        self.actionMark_QSL_received.setIcon(icon28)
        self.actionMark_QSL_received.setObjectName("actionMark_QSL_received")
        self.actionMark_HamQTH_uploaded = QtGui.QAction(parent=MainWindow)
        self.actionMark_HamQTH_uploaded.setIcon(icon28)
        self.actionMark_HamQTH_uploaded.setObjectName("actionMark_HamQTH_uploaded")
        self.actionUnmark_QSL_sent = QtGui.QAction(parent=MainWindow)
        self.actionUnmark_QSL_sent.setIcon(icon1)
        self.actionUnmark_QSL_sent.setObjectName("actionUnmark_QSL_sent")
        self.actionUnmark_QSL_received = QtGui.QAction(parent=MainWindow)
        self.actionUnmark_QSL_received.setIcon(icon1)
        self.actionUnmark_QSL_received.setObjectName("actionUnmark_QSL_received")
        self.actionShow_Contest_stats = QtGui.QAction(parent=MainWindow)
        icon29 = QtGui.QIcon()
        icon29.addPixmap(QtGui.QPixmap("icons:contest_stats.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionShow_Contest_stats.setIcon(icon29)
        self.actionShow_Contest_stats.setObjectName("actionShow_Contest_stats")
        self.actionShow_statistics = QtGui.QAction(parent=MainWindow)
        self.actionShow_statistics.setEnabled(False)
        icon30 = QtGui.QIcon()
        icon30.addPixmap(QtGui.QPixmap("icons:log.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionShow_statistics.setIcon(icon30)
        self.actionShow_statistics.setObjectName("actionShow_statistics")
        self.actionShow_statistics_TB = QtGui.QAction(parent=MainWindow)
        self.actionShow_statistics_TB.setEnabled(False)
        self.actionShow_statistics_TB.setIcon(icon30)
        self.actionShow_statistics_TB.setObjectName("actionShow_statistics_TB")
        self.actionSelect_callbook = QtGui.QAction(parent=MainWindow)
        icon31 = QtGui.QIcon()
        icon31.addPixmap(QtGui.QPixmap("icons:contents.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionSelect_callbook.setIcon(icon31)
        self.actionSelect_callbook.setObjectName("actionSelect_callbook")
        self.actionRefresh_callbook = QtGui.QAction(parent=MainWindow)
        self.actionRefresh_callbook.setEnabled(False)
        icon32 = QtGui.QIcon()
        icon32.addPixmap(QtGui.QPixmap("icons:refresh_callbook.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionRefresh_callbook.setIcon(icon32)
        self.actionRefresh_callbook.setObjectName("actionRefresh_callbook")
        self.actionRefresh_call_history = QtGui.QAction(parent=MainWindow)
        self.actionRefresh_call_history.setEnabled(False)
        icon33 = QtGui.QIcon()
        icon33.addPixmap(QtGui.QPixmap("icons:refresh_history.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionRefresh_call_history.setIcon(icon33)
        self.actionRefresh_call_history.setObjectName("actionRefresh_call_history")
        self.actionExport_callbook = QtGui.QAction(parent=MainWindow)
        self.actionExport_callbook.setEnabled(False)
        icon34 = QtGui.QIcon()
        icon34.addPixmap(QtGui.QPixmap("icons:callbook_export.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionExport_callbook.setIcon(icon34)
        self.actionExport_callbook.setObjectName("actionExport_callbook")
        self.actionImport_callbook = QtGui.QAction(parent=MainWindow)
        self.actionImport_callbook.setEnabled(False)
        icon35 = QtGui.QIcon()
        icon35.addPixmap(QtGui.QPixmap("icons:callbook_import.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionImport_callbook.setIcon(icon35)
        self.actionImport_callbook.setObjectName("actionImport_callbook")
        self.actionExport_call_history = QtGui.QAction(parent=MainWindow)
        self.actionExport_call_history.setEnabled(False)
        icon36 = QtGui.QIcon()
        icon36.addPixmap(QtGui.QPixmap("icons:history_export.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionExport_call_history.setIcon(icon36)
        self.actionExport_call_history.setObjectName("actionExport_call_history")
        self.actionImport_call_history = QtGui.QAction(parent=MainWindow)
        self.actionImport_call_history.setEnabled(False)
        icon37 = QtGui.QIcon()
        icon37.addPixmap(QtGui.QPixmap("icons:history_import.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionImport_call_history.setIcon(icon37)
        self.actionImport_call_history.setObjectName("actionImport_call_history")
        self.actionCopy = QtGui.QAction(parent=MainWindow)
        self.actionCopy.setEnabled(False)
        icon38 = QtGui.QIcon()
        icon38.addPixmap(QtGui.QPixmap("icons:editcopy.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionCopy.setIcon(icon38)
        self.actionCopy.setObjectName("actionCopy")
        self.actionRead_DARC_QSL_QR_Code = QtGui.QAction(parent=MainWindow)
        self.actionRead_DARC_QSL_QR_Code.setEnabled(False)
        icon39 = QtGui.QIcon()
        icon39.addPixmap(QtGui.QPixmap("icons:icons8-qrcode-96.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionRead_DARC_QSL_QR_Code.setIcon(icon39)
        self.actionRead_DARC_QSL_QR_Code.setObjectName("actionRead_DARC_QSL_QR_Code")
        self.actionRead_DARC_QSL_QR_Code_TB = QtGui.QAction(parent=MainWindow)
        self.actionRead_DARC_QSL_QR_Code_TB.setEnabled(False)
        self.actionRead_DARC_QSL_QR_Code_TB.setIcon(icon39)
        self.actionRead_DARC_QSL_QR_Code_TB.setObjectName("actionRead_DARC_QSL_QR_Code_TB")
        self.actionChange_contest = QtGui.QAction(parent=MainWindow)
        icon40 = QtGui.QIcon()
        icon40.addPixmap(QtGui.QPixmap("icons:contest_edit.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.actionChange_contest.setIcon(icon40)
        self.actionChange_contest.setObjectName("actionChange_contest")
        self.menuFile.addAction(self.actionSelect_DB)
        self.menuFile.addAction(self.actionExport)
        self.menuFile.addAction(self.actionImport)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionSelect_callbook)
        self.menuFile.addAction(self.actionRefresh_callbook)
        self.menuFile.addAction(self.actionExport_callbook)
        self.menuFile.addAction(self.actionImport_callbook)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionSettings)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionExit)
        self.menuEdit.addAction(self.actionLog_QSO)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionDelete_log_entry)
        self.menuEdit.addAction(self.actionChange_log_entry)
        self.menuEdit.addAction(self.actionCopy)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionMark_QSL_sent)
        self.menuEdit.addAction(self.actionUnmark_QSL_sent)
        self.menuEdit.addAction(self.actionMark_QSL_received)
        self.menuEdit.addAction(self.actionUnmark_QSL_received)
        self.menuEdit.addAction(self.actionMark_HamQTH_uploaded)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionChange_contest)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionWatch_file_for_QSOs)
        self.menuHelp.addAction(self.actionHelp)
        self.menuHelp.addAction(self.actionHelpCassiopeiaConsole)
        self.menuHelp.addAction(self.actionShortcuts)
        self.menuHelp.addSeparator()
        self.menuHelp.addAction(self.actionShow_CassiopeiaConsole)
        self.menuHelp.addAction(self.actionShow_filter)
        self.menuHelp.addAction(self.actionShow_DX_Spots)
        self.menuHelp.addAction(self.actionShow_log)
        self.menuHelp.addAction(self.actionShow_statistics)
        self.menuHelp.addSeparator()
        self.menuHelp.addAction(self.actionAbout)
        self.menuHelp.addAction(self.actionAbout_Qt)
        self.menuUpload.addAction(self.actionUpload_to_eQSL)
        self.menuUpload.addAction(self.actionCheck_eQSL)
        self.menuUpload.addAction(self.actionDownload_eQSL)
        self.menuUpload.addSeparator()
        self.menuUpload.addAction(self.actionUpload_logs_to_LoTW)
        self.menuUpload.addAction(self.actionCheck_LoTW_QSL)
        self.menuUpload.addSeparator()
        self.menuUpload.addAction(self.actionUpload_to_HamQTH)
        self.menuUpload.addSeparator()
        self.menuUpload.addAction(self.actionRead_DARC_QSL_QR_Code)
        self.menuContest.addAction(self.actionShow_Contest_stats)
        self.menuContest.addAction(self.actionExport_Contest_Log)
        self.menuContest.addSeparator()
        self.menuContest.addAction(self.actionContests)
        self.menuContest.addSeparator()
        self.menuContest.addAction(self.actionRefresh_call_history)
        self.menuContest.addAction(self.actionExport_call_history)
        self.menuContest.addAction(self.actionImport_call_history)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuEdit.menuAction())
        self.menubar.addAction(self.menuUpload.menuAction())
        self.menubar.addAction(self.menuContest.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())
        self.toolBar.addAction(self.actionSelect_DB_TB)
        self.toolBar.addAction(self.actionImport_TB)
        self.toolBar.addAction(self.actionExport_TB)
        self.toolBar.addAction(self.actionSettings_TB)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.actionStart_hamlib_TB)
        self.toolBar.addAction(self.actionLog_QSO_TB)
        self.toolBar.addAction(self.actionWatch_file_for_QSOs_TB)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.actionShow_statistics_TB)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.actionRead_DARC_QSL_QR_Code_TB)

        self.retranslateUi(MainWindow)
        self.actionAbout.triggered.connect(MainWindow.showAbout) # type: ignore
        self.actionAbout_Qt.triggered.connect(MainWindow.showAboutQt) # type: ignore
        self.actionLog_QSO.triggered.connect(MainWindow.logQSO) # type: ignore
        self.actionLog_QSO_TB.triggered.connect(MainWindow.logQSO) # type: ignore
        self.actionExit.triggered.connect(MainWindow.close) # type: ignore
        self.actionSelect_DB.triggered.connect(MainWindow.selectDB) # type: ignore
        self.actionDelete_log_entry.triggered.connect(MainWindow.deleteQSO) # type: ignore
        self.actionChange_log_entry.triggered.connect(MainWindow.changeQSO) # type: ignore
        self.actionSettings.triggered.connect(MainWindow.showSettings) # type: ignore
        self.actionSettings_TB.triggered.connect(MainWindow.showSettings) # type: ignore
        self.actionExport.triggered.connect(MainWindow.export) # type: ignore
        self.actionExport_TB.triggered.connect(MainWindow.export) # type: ignore
        self.QSOTableView.doubleClicked['QModelIndex'].connect(MainWindow.changeQSO) # type: ignore
        self.actionHelp.triggered.connect(MainWindow.showHelp) # type: ignore
        self.actionImport.triggered.connect(MainWindow.logImport) # type: ignore
        self.actionImport_TB.triggered.connect(MainWindow.logImport) # type: ignore
        self.actionStart_hamlib_TB.triggered['bool'].connect(MainWindow.ctrlHamlib) # type: ignore
        self.actionSelect_DB_TB.triggered.connect(MainWindow.selectDB) # type: ignore
        self.actionWatch_file_for_QSOs.triggered['bool'].connect(MainWindow.ctrlWatching) # type: ignore
        self.actionWatch_file_for_QSOs_TB.triggered['bool'].connect(MainWindow.ctrlWatching) # type: ignore
        self.actionShow_log.triggered.connect(self.logDockWidget.show) # type: ignore
        self.actionUpload_logs_to_LoTW.triggered.connect(MainWindow.lotwUpload) # type: ignore
        self.QSOTableView.clicked['QModelIndex'].connect(MainWindow.clearQSOForm) # type: ignore
        self.resetFilterPushButton.clicked.connect(MainWindow.resetTableFilter) # type: ignore
        self.actionShow_filter.triggered.connect(self.filterDockWidget.show) # type: ignore
        self.fCallsignLineEdit.editingFinished.connect(MainWindow.setTableFilter) # type: ignore
        self.fBandComboBox.currentTextChanged['QString'].connect(MainWindow.setTableFilter) # type: ignore
        self.fModeComboBox.currentTextChanged['QString'].connect(MainWindow.setTableFilter) # type: ignore
        self.fDateToLineEdit.editingFinished.connect(MainWindow.setTableFilter) # type: ignore
        self.fDateFromLineEdit.editingFinished.connect(MainWindow.setTableFilter) # type: ignore
        self.actionShow_CassiopeiaConsole.triggered.connect(MainWindow.showCC) # type: ignore
        self.actionUpload_to_HamQTH.triggered.connect(MainWindow.uploadToHamQTH) # type: ignore
        self.actionCheck_LoTW_QSL.triggered.connect(MainWindow.lotwCheckQSL) # type: ignore
        self.actionCheck_eQSL.triggered.connect(MainWindow.eqslCheckInboxSelected) # type: ignore
        self.actionDownload_eQSL.triggered.connect(MainWindow.eqslDownload) # type: ignore
        self.actionUpload_to_eQSL.triggered.connect(MainWindow.eqslUpload) # type: ignore
        self.actionShortcuts.triggered.connect(MainWindow.showShortcuts) # type: ignore
        self.actionHelpCassiopeiaConsole.triggered.connect(MainWindow.showCCHelp) # type: ignore
        self.actionShow_DX_Spots.triggered.connect(self.dxSpotsDockWidget.show) # type: ignore
        self.actionExport_Contest_Log.triggered.connect(MainWindow.exportContestLog) # type: ignore
        self.actionContests.triggered.connect(MainWindow.showContestHelp) # type: ignore
        self.fQSLsComboBox.currentIndexChanged['int'].connect(MainWindow.setTableFilter) # type: ignore
        self.fQSLrComboBox.currentIndexChanged['int'].connect(MainWindow.setTableFilter) # type: ignore
        self.feQSLsComboBox.currentIndexChanged['int'].connect(MainWindow.setTableFilter) # type: ignore
        self.feQSLrComboBox.currentIndexChanged['int'].connect(MainWindow.setTableFilter) # type: ignore
        self.fLoTWsComboBox.currentIndexChanged['int'].connect(MainWindow.setTableFilter) # type: ignore
        self.fLoTWrComboBox.currentIndexChanged['int'].connect(MainWindow.setTableFilter) # type: ignore
        self.fContestComboBox.currentTextChanged['QString'].connect(MainWindow.setTableFilter) # type: ignore
        self.actionMark_HamQTH_uploaded.triggered.connect(MainWindow.markQSO) # type: ignore
        self.actionMark_QSL_received.triggered.connect(MainWindow.markQSO) # type: ignore
        self.actionMark_QSL_sent.triggered.connect(MainWindow.markQSO) # type: ignore
        self.actionUnmark_QSL_received.triggered.connect(MainWindow.markQSO) # type: ignore
        self.actionUnmark_QSL_sent.triggered.connect(MainWindow.markQSO) # type: ignore
        self.actionShow_Contest_stats.triggered.connect(self.contestStatDockWidget.show) # type: ignore
        self.actionShow_statistics.triggered.connect(MainWindow.showStatistics) # type: ignore
        self.actionShow_statistics_TB.triggered.connect(MainWindow.showStatistics) # type: ignore
        self.actionSelect_callbook.triggered.connect(MainWindow.selectCallbook) # type: ignore
        self.actionRefresh_call_history.triggered.connect(MainWindow.initialiseCallHistory) # type: ignore
        self.actionRefresh_callbook.triggered.connect(MainWindow.initialiseCallbook) # type: ignore
        self.actionExport_call_history.triggered.connect(MainWindow.exportCallHistory) # type: ignore
        self.actionExport_callbook.triggered.connect(MainWindow.exportCallbook) # type: ignore
        self.actionImport_call_history.triggered.connect(MainWindow.importCallHistory) # type: ignore
        self.actionImport_callbook.triggered.connect(MainWindow.importCallbook) # type: ignore
        self.actionCopy.triggered.connect(MainWindow.copyToClipboard) # type: ignore
        self.actionRead_DARC_QSL_QR_Code_TB.triggered.connect(MainWindow.readQRCode) # type: ignore
        self.actionWatch_file_for_QSOs.triggered.connect(MainWindow.readQRCode) # type: ignore
        self.actionChange_contest.triggered.connect(MainWindow.changeContestID) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "DragonLog"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.menuEdit.setTitle(_translate("MainWindow", "Edit"))
        self.menuHelp.setTitle(_translate("MainWindow", "Help"))
        self.menuUpload.setTitle(_translate("MainWindow", "Upload && QSL"))
        self.menuContest.setTitle(_translate("MainWindow", "Contest"))
        self.toolBar.setWindowTitle(_translate("MainWindow", "Toolbar"))
        self.logDockWidget.setWindowTitle(_translate("MainWindow", "Application Log"))
        self.qsoDockWidget.setWindowTitle(_translate("MainWindow", "QSO Form"))
        self.filterDockWidget.setWindowTitle(_translate("MainWindow", "Filter"))
        self.label_4.setText(_translate("MainWindow", "Date from"))
        self.fDateFromLineEdit.setPlaceholderText(_translate("MainWindow", "YYYY-MM-DD"))
        self.label_5.setText(_translate("MainWindow", "Date to"))
        self.fDateToLineEdit.setPlaceholderText(_translate("MainWindow", "YYYY-MM-DD"))
        self.label.setText(_translate("MainWindow", "Callsign"))
        self.label_3.setText(_translate("MainWindow", "Band"))
        self.label_2.setText(_translate("MainWindow", "Mode"))
        self.label_6.setText(_translate("MainWindow", "QSL sent/rcvd"))
        self.fQSLsComboBox.setItemText(1, _translate("MainWindow", "Yes"))
        self.fQSLsComboBox.setItemText(2, _translate("MainWindow", "No"))
        self.fQSLrComboBox.setItemText(1, _translate("MainWindow", "Yes"))
        self.fQSLrComboBox.setItemText(2, _translate("MainWindow", "No"))
        self.label_7.setText(_translate("MainWindow", "eQSL  sent/rcvd"))
        self.feQSLsComboBox.setItemText(1, _translate("MainWindow", "Yes"))
        self.feQSLsComboBox.setItemText(2, _translate("MainWindow", "No"))
        self.feQSLrComboBox.setItemText(1, _translate("MainWindow", "Yes"))
        self.feQSLrComboBox.setItemText(2, _translate("MainWindow", "No"))
        self.label_8.setText(_translate("MainWindow", "LoTW  sent/rcvd"))
        self.fLoTWsComboBox.setItemText(1, _translate("MainWindow", "Yes"))
        self.fLoTWsComboBox.setItemText(2, _translate("MainWindow", "No"))
        self.fLoTWrComboBox.setItemText(1, _translate("MainWindow", "Yes"))
        self.fLoTWrComboBox.setItemText(2, _translate("MainWindow", "No"))
        self.label_9.setText(_translate("MainWindow", "Contest"))
        self.resetFilterPushButton.setShortcut(_translate("MainWindow", "Ctrl+R"))
        self.ccDockWidget.setWindowTitle(_translate("MainWindow", "CassiopeiaConsole"))
        self.dxSpotsDockWidget.setWindowTitle(_translate("MainWindow", "DX Spots"))
        self.contestStatDockWidget.setWindowTitle(_translate("MainWindow", "Contest Statistics"))
        self.actionLog_QSO.setText(_translate("MainWindow", "Log QSO..."))
        self.actionLog_QSO.setShortcut(_translate("MainWindow", "Ctrl+L"))
        self.actionLog_QSO_TB.setText(_translate("MainWindow", "Log QSO..."))
        self.actionSelect_DB.setText(_translate("MainWindow", "Select database..."))
        self.actionSelect_DB_TB.setText(_translate("MainWindow", "Select database..."))
        self.actionExit.setText(_translate("MainWindow", "Exit"))
        self.actionExit.setShortcut(_translate("MainWindow", "Ctrl+Q"))
        self.actionExport.setText(_translate("MainWindow", "Export..."))
        self.actionImport.setText(_translate("MainWindow", "Import..."))
        self.actionExport_TB.setText(_translate("MainWindow", "Export"))
        self.actionImport_TB.setText(_translate("MainWindow", "Import"))
        self.actionAbout.setText(_translate("MainWindow", "About"))
        self.actionAbout_Qt.setText(_translate("MainWindow", "About Qt"))
        self.actionDelete_log_entry.setText(_translate("MainWindow", "Delete log entry"))
        self.actionDelete_log_entry.setShortcut(_translate("MainWindow", "Ctrl+X"))
        self.actionChange_log_entry.setText(_translate("MainWindow", "Change log entry..."))
        self.actionChange_log_entry.setShortcut(_translate("MainWindow", "Ctrl+E"))
        self.actionSettings.setText(_translate("MainWindow", "Settings"))
        self.actionSettings_TB.setText(_translate("MainWindow", "Settings"))
        self.actionHelp.setText(_translate("MainWindow", "Help"))
        self.actionStart_hamlib_TB.setText(_translate("MainWindow", "Start hamlib"))
        self.actionWatch_file_for_QSOs.setText(_translate("MainWindow", "Watch file for QSOs..."))
        self.actionWatch_file_for_QSOs.setShortcut(_translate("MainWindow", "Ctrl+W"))
        self.actionWatch_file_for_QSOs_TB.setText(_translate("MainWindow", "Watch file for QSOs..."))
        self.actionShow_log.setText(_translate("MainWindow", "Show log"))
        self.actionUpload_logs_to_LoTW.setText(_translate("MainWindow", "Upload logs to LoTW..."))
        self.actionShow_filter.setText(_translate("MainWindow", "Show filter"))
        self.actionShow_filter.setShortcut(_translate("MainWindow", "Ctrl+F"))
        self.actionShow_CassiopeiaConsole.setText(_translate("MainWindow", "Show CassiopeiaConsole"))
        self.actionShow_CassiopeiaConsole.setShortcut(_translate("MainWindow", "Ctrl+K"))
        self.actionUpload_to_HamQTH.setText(_translate("MainWindow", "Upload to HamQTH"))
        self.actionCheck_LoTW_QSL.setText(_translate("MainWindow", "Check LoTW QSL"))
        self.actionCheck_LoTW_QSL.setShortcut(_translate("MainWindow", "Ctrl+Alt+L"))
        self.actionUpload_to_eQSL.setText(_translate("MainWindow", "Upload to eQSL"))
        self.actionCheck_eQSL.setText(_translate("MainWindow", "Check eQSL"))
        self.actionCheck_eQSL.setShortcut(_translate("MainWindow", "Ctrl+Alt+E"))
        self.actionDownload_eQSL.setText(_translate("MainWindow", "Download eQSL"))
        self.actionDownload_eQSL.setShortcut(_translate("MainWindow", "Ctrl+Shift+E"))
        self.actionShortcuts.setText(_translate("MainWindow", "Shortcuts"))
        self.actionHelpCassiopeiaConsole.setText(_translate("MainWindow", "Help on CassiopeiaConsole"))
        self.actionShow_DX_Spots.setText(_translate("MainWindow", "Show DX Spots"))
        self.actionShow_DX_Spots.setShortcut(_translate("MainWindow", "Ctrl+D"))
        self.actionExport_Contest_Log.setText(_translate("MainWindow", "Export Contest Log..."))
        self.actionContests.setText(_translate("MainWindow", "Contests"))
        self.actionMark_QSL_sent.setText(_translate("MainWindow", "Mark QSL sent"))
        self.actionMark_QSL_sent.setShortcut(_translate("MainWindow", "Ctrl+Alt+S"))
        self.actionMark_QSL_received.setText(_translate("MainWindow", "Mark QSL received"))
        self.actionMark_QSL_received.setShortcut(_translate("MainWindow", "Ctrl+Alt+R"))
        self.actionMark_HamQTH_uploaded.setText(_translate("MainWindow", "Mark HamQTH uploaded"))
        self.actionMark_HamQTH_uploaded.setShortcut(_translate("MainWindow", "Ctrl+Alt+U"))
        self.actionUnmark_QSL_sent.setText(_translate("MainWindow", "Unmark QSL sent"))
        self.actionUnmark_QSL_sent.setShortcut(_translate("MainWindow", "Ctrl+Shift+S"))
        self.actionUnmark_QSL_received.setText(_translate("MainWindow", "Unmark QSL received"))
        self.actionUnmark_QSL_received.setShortcut(_translate("MainWindow", "Ctrl+Shift+R"))
        self.actionShow_Contest_stats.setText(_translate("MainWindow", "Show Contest stats"))
        self.actionShow_Contest_stats.setShortcut(_translate("MainWindow", "Ctrl+T"))
        self.actionShow_statistics.setText(_translate("MainWindow", "Show statistics"))
        self.actionShow_statistics_TB.setText(_translate("MainWindow", "Show statistics"))
        self.actionSelect_callbook.setText(_translate("MainWindow", "Select callbook..."))
        self.actionRefresh_callbook.setText(_translate("MainWindow", "Refresh callbook"))
        self.actionRefresh_call_history.setText(_translate("MainWindow", "Refresh call history"))
        self.actionExport_callbook.setText(_translate("MainWindow", "Export callbook..."))
        self.actionImport_callbook.setText(_translate("MainWindow", "Import callbook..."))
        self.actionExport_call_history.setText(_translate("MainWindow", "Export call history..."))
        self.actionImport_call_history.setText(_translate("MainWindow", "Import call history..."))
        self.actionCopy.setText(_translate("MainWindow", "Copy"))
        self.actionCopy.setShortcut(_translate("MainWindow", "Ctrl+C"))
        self.actionRead_DARC_QSL_QR_Code.setText(_translate("MainWindow", "Read DARC QSL-QR-Code..."))
        self.actionRead_DARC_QSL_QR_Code_TB.setText(_translate("MainWindow", "Read DARC QSL-QR-Code"))
        self.actionChange_contest.setText(_translate("MainWindow", "Change contest..."))
        self.actionChange_contest.setShortcut(_translate("MainWindow", "Ctrl+Alt+C"))



# DragonLog (c) 2025 by Andreas Schawo is licensed under CC BY-SA 4.0.
# To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/4.0/

from PyQt6 import QtWidgets, QtSql, QtGui
import numpy as np
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg, NavigationToolbar2QT
from matplotlib.figure import Figure


# noinspection PyPep8Naming
class StatisticsWidget(QtWidgets.QDialog):
    def __init__(self, parent, title: str, db_con: QtSql.QSqlDatabase, bands: tuple):
        super().__init__(parent)

        self.__db_con__ = db_con
        self.__bands__ = bands

        self.initTemporary()

        self.setWindowTitle(title)
        self.setModal(True)
        self.setWhatsThis(self.tr('Press Ctrl+C to copy the data to the clipboard'))
        verticalLayout = QtWidgets.QVBoxLayout()
        self.setLayout(verticalLayout)
        self.tabWidget = QtWidgets.QTabWidget()
        verticalLayout.addWidget(self.tabWidget)

        self.buildTableStat('QSO', (self.tr('QSOs'), self.tr('Modes'), self.tr('Bands'),
                                    self.tr('Gridsquares'), self.tr('Contest QSOs')))
        self.buildQSLStat('QSL', (self.tr('QSOs'),
                                  self.tr('QSL sent'), self.tr('QSL rcvd'),
                                  self.tr('eQSL sent'), self.tr('eQSL rcvd'),
                                  self.tr('LotW sent'), self.tr('LotW rcvd')))
        self.buildGraphStat(self.tr('Modes'), 'mode')
        self.buildGraphStat(self.tr('Bands'), 'band')

        self.show()

    def initTemporary(self):
        # QSO stat
        self.__db_con__.exec('''CREATE TEMP VIEW IF NOT EXISTS t_qso_stat AS
            SELECT strftime('%Y', date_time) as year, COUNT(id) as qsos, COUNT(DISTINCT mode) as modes, 
                COUNT(DISTINCT band) as bands, COUNT(DISTINCT upper(substr(locator, 0, 4))) as gridsquares, 
                COUNT(nullif(contest_id = '', 1)) as contest_qsos
            FROM qsos GROUP BY year
            UNION SELECT 'total', COUNT(id), COUNT(DISTINCT mode), COUNT(DISTINCT band), 
                COUNT(DISTINCT upper(substr(locator, 0, 4))), COUNT(nullif(contest_id = '', 1)) 
            FROM qsos;''')
        # QSL stat
        self.__db_con__.exec('''CREATE TEMP VIEW IF NOT EXISTS t_qsl_stat AS 
            SELECT
                strftime('%Y', date_time) as year, COUNT(id) as qsos, 
                COUNT(nullif(qsl_sent != 'Y', 1)) as qsl_sent, COUNT(nullif(qsl_rcvd != 'Y', 1)) as qsl_rcvd,
                COUNT(nullif(eqsl_sent != 'Y', 1)) as eqsl_sent, COUNT(nullif(eqsl_rcvd != 'Y', 1)) as eqsl_rcvd,
                COUNT(nullif(lotw_sent != 'Y', 1)) as lotw_sent, COUNT(nullif(lotw_rcvd != 'Y', 1)) as lotw_rcvd
            FROM qsos GROUP BY year
            UNION SELECT 'total', COUNT(id) as qsos, 
                COUNT(nullif(qsl_sent != 'Y', 1)), COUNT(nullif(qsl_rcvd != 'Y', 1)),
                COUNT(nullif(eqsl_sent != 'Y', 1)), COUNT(nullif(eqsl_rcvd != 'Y', 1)),
                COUNT(nullif(lotw_sent != 'Y', 1)), COUNT(nullif(lotw_rcvd != 'Y', 1)) 
            FROM qsos;''')

        # Bands table for proper sorting
        query = self.__db_con__.exec('''SELECT count(*) FROM sqlite_master WHERE type='table' AND name='t_bands';''')
        if query.next() and query.value(0) == 0:
            self.__db_con__.exec('''CREATE TEMP TABLE IF NOT EXISTS t_bands (
                    "id"    INTEGER NOT NULL,
                    "band"  TEXT,
                    PRIMARY KEY("id" AUTOINCREMENT)
                );''')
            for b in self.__bands__:
                self.__db_con__.exec(f"INSERT INTO t_bands (band) VALUES ('{b}')")

        # Bands summary
        self.__db_con__.exec('''CREATE TEMP VIEW IF NOT EXISTS t_band_stat AS 
            SELECT band, COUNT(*) as qsos FROM qsos GROUP BY band;''')
        # Modes summary
        self.__db_con__.exec('''CREATE TEMP VIEW IF NOT EXISTS t_mode_stat AS 
            SELECT mode, COUNT(*) as qsos FROM qsos GROUP BY mode;''')
        # Bands per year
        self.__db_con__.exec('''CREATE TEMP VIEW IF NOT EXISTS t_bands_per_year AS 
            WITH 
                years AS (SELECT DISTINCT strftime('%Y', date_time) as year FROM qsos),
                dist_bands AS (SELECT DISTINCT band FROM qsos),
                year_band AS (SELECT year, band FROM years CROSS JOIN dist_bands)
            SELECT 
                yb.year,
                yb.band,
                COALESCE(COUNT(q.id), 0) AS qso_count
            FROM 
                year_band yb
            LEFT JOIN 
                qsos q ON strftime('%Y', q.date_time) = yb.year AND q.band = yb.band
            LEFT JOIN 
                t_bands b ON b.band = yb.band 
            GROUP BY 
                yb.year, yb.band
            ORDER BY 
                yb.year, b.id;
            ''')
        # Modes per year
        self.__db_con__.exec('''CREATE TEMP VIEW IF NOT EXISTS t_modes_per_year AS 
            WITH 
                years AS (SELECT DISTINCT strftime('%Y', date_time) as year FROM qsos),
                modes AS (SELECT DISTINCT mode FROM qsos),
                year_mode AS (SELECT year, mode FROM years CROSS JOIN modes)
            SELECT 
                ym.year,
                ym.mode,
                COALESCE(COUNT(q.id), 0) AS qso_count
            FROM 
                year_mode ym
            LEFT JOIN 
                qsos q ON strftime('%Y', q.date_time) = ym.year AND q.mode = ym.mode
            GROUP BY 
                ym.year, ym.mode
            ORDER BY 
                ym.year, ym.mode;
            ''')
        self.__db_con__.commit()

    def fetchStat(self, name='qso') -> list[list]:
        stat = []
        if self.__db_con__ and name in ('qso', 'qsl', 'band', 'mode'):
            query = self.__db_con__.exec(f'SELECT * FROM t_{name}_stat')
            if query.lastError().text():
                raise Exception(query.lastError().text())

            rec = query.record()
            col_count = rec.count()
            stat.append([rec.fieldName(i) for i in range(col_count)])
            while query.next():
                stat.append([query.value(i) for i in range(col_count)])
        return stat

    def fetchPerYearStat(self, name) -> dict[str, list]:
        stat: dict[str, list] = {
            'categories': [],
        }
        if self.__db_con__ and name in ('band', 'mode'):
            query = self.__db_con__.exec(f'SELECT * FROM t_{name}s_per_year')
            if query.lastError().text():
                raise Exception(query.lastError().text())
            while query.next():
                if query.value(1) not in stat['categories']:
                    stat['categories'].append(query.value(1))
                if query.value(0) not in stat:
                    stat[query.value(0)] = [query.value(2)]
                else:
                    stat[query.value(0)].append(query.value(2))
        return stat

    def buildTableStat(self, title: str, header: tuple = ()):
        stat = self.fetchStat(title.lower())
        if not stat:
            return

        tableWidget = StatTableWidget(len(stat) - 1, len(stat[0]) - 1, stat_summary=stat.copy())
        tableWidget.setEditTriggers(QtWidgets.QAbstractItemView.EditTrigger.NoEditTriggers)
        tableWidget.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.NoSelection)
        tableWidget.setVerticalHeaderLabels([str(r[0]).capitalize() for r in stat[1:]])

        if header:
            tableWidget.setHorizontalHeaderLabels(header)
        else:
            tableWidget.setHorizontalHeaderLabels([hl.capitalize() for hl in stat[0][1:]])

        self.tabWidget.addTab(tableWidget, f'{title}-{self.tr("Statistic")}')

        for r, row in enumerate(stat[1:]):
            for c, cell in enumerate(row[1:]):
                item = QtWidgets.QTableWidgetItem(str(cell))
                tableWidget.setItem(r, c, item)

        tableWidget.resizeColumnsToContents()

    def buildQSLStat(self, title: str, header: tuple = ()):
        stat = self.fetchStat(title.lower())
        if not stat:
            return

        widget = StatWidget(self, stat_summary=stat)

        # Table
        tableWidget = StatTableWidget(len(stat) - 1, len(stat[0]) - 1, stat_summary=stat.copy())
        tableWidget.setEditTriggers(QtWidgets.QAbstractItemView.EditTrigger.NoEditTriggers)
        tableWidget.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.NoSelection)
        tableWidget.setVerticalHeaderLabels([str(r[0]).capitalize() for r in stat[1:]])

        if header:
            tableWidget.setHorizontalHeaderLabels(header)
        else:
            tableWidget.setHorizontalHeaderLabels([hl.capitalize() for hl in stat[0][1:]])

        for r, row in enumerate(stat[1:]):
            for c, cell in enumerate(row[1:]):
                item = QtWidgets.QTableWidgetItem(str(cell))
                tableWidget.setItem(r, c, item)
        tableWidget.resizeColumnsToContents()

        # Graph
        stat_widget = FigureCanvasQTAgg(Figure())
        ax = stat_widget.figure.subplots()
        ax.tick_params(axis='x', rotation=90)
        cat = header
        width = .8
        bottom = np.zeros(len(header))

        for d in stat[1:-1]:
            val = [int(v) for v in d[1:]]
            p = ax.bar(cat, val, width, label=d[0], bottom=bottom)
            bottom += val
        stat_widget.figure.tight_layout()

        # Layout
        layout = QtWidgets.QVBoxLayout(widget)
        layout.addWidget(tableWidget)
        layout.addWidget(stat_widget)
        layout.addWidget(NavigationToolbar(stat_widget, self.parent()))
        self.tabWidget.addTab(widget, f'{title}-{self.tr("Statistic")}')

    def buildGraphStat(self, title: str, stat_name: str):
        stat_sum = self.fetchStat(stat_name)
        if not stat_sum:
            return

        stat_detail = self.fetchPerYearStat(stat_name)
        if not stat_detail:
            return

        # Init
        widget = StatWidget(self, stat_summary=stat_sum.copy(), stat_details=stat_detail.copy())
        stat_widget = FigureCanvasQTAgg(Figure())
        ax = stat_widget.figure.subplots(1, 2)

        # per year
        stat_cat = stat_detail.pop('categories')
        x = np.arange(len(stat_cat))
        width = 1 / (len(stat_detail) + 1)  # the width of the bars

        try:
            for i, (cat, count) in enumerate(stat_detail.items()):
                offset = width * i
                rects = ax[0].bar(x + offset, count, width, label=cat)
                # ax[0].bar_label(rects, padding=3)
                ax[0].tick_params(axis='x', rotation=90)

            ax[0].set_ylabel('QSOs')
            ax[0].set_title(f'{title} ' + self.tr('per year'))
            ax[0].set_xticks(x + width, stat_cat)
            ax[0].legend(fontsize='small')

            # summary
            stat_sum = np.array(stat_sum[1:]).transpose()
            ax[1].pie([int(c) for c in stat_sum[1]], labels=stat_sum[0])
            ax[1].set_title(f'{title} ' + self.tr('summary'))

            # display
            stat_widget.figure.tight_layout()
            layout = QtWidgets.QVBoxLayout(widget)
            layout.addWidget(stat_widget)
            layout.addWidget(NavigationToolbar(stat_widget, self.parent()))
            self.tabWidget.addTab(widget, f'{title}-{self.tr("Statistic")}')
        except IndexError:
            pass

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        if type(event) is QtGui.QKeyEvent:
            if event.matches(QtGui.QKeySequence.StandardKey.Copy):
                # noinspection PyTypeChecker
                widget: StatWidget = self.tabWidget.currentWidget()
                table = []
                if widget.stat_details:
                    for r, d in widget.stat_details.items():
                        d.insert(0, r)
                        table.append('\t'.join([str(c) for c in d]))
                else:
                    for r in widget.stat_summary:
                        table.append('\t'.join([str(c) for c in r]))

                clip = QtWidgets.QApplication.clipboard()
                clip.setText('\r\n'.join(table))


class NavigationToolbar(NavigationToolbar2QT):
    toolitems = [t for t in NavigationToolbar2QT.toolitems if
                 t[0] in ('Home', 'Pan', 'Zoom', 'Save')]


class StatWidget(QtWidgets.QWidget):
    def __init__(self, *args, stat_summary: list[list], stat_details: dict[str, list] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.__stat_summary__: list[list] = stat_summary
        self.__stat_details__: dict[str, list] = stat_details

    @property
    def stat_summary(self) -> list[list]:
        return self.__stat_summary__

    @property
    def stat_details(self) -> dict[str, list]:
        return self.__stat_details__


class StatTableWidget(QtWidgets.QTableWidget, StatWidget):
    def __init__(self, *args, stat_summary: list[list], stat_details: dict[str, list] = None, **kwargs):
        super().__init__(*args, stat_summary=stat_summary, stat_details=stat_details, **kwargs)
        self.__stat_summary__: list[list] = stat_summary
        self.__stat_details__: dict[str, list] = stat_details



# Form implementation generated from reading ui file 'ContestStatistics.ui'
#
# Created by: PyQt6 UI code generator 6.9.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_ContestStatistics(object):
    def setupUi(self, ContestStatistics):
        ContestStatistics.setObjectName("ContestStatistics")
        ContestStatistics.resize(350, 370)
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(ContestStatistics)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label_3 = QtWidgets.QLabel(parent=ContestStatistics)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_3.addWidget(self.label_3)
        self.contestComboBox = QtWidgets.QComboBox(parent=ContestStatistics)
        self.contestComboBox.setObjectName("contestComboBox")
        self.verticalLayout_3.addWidget(self.contestComboBox)
        self.horizontalLayout.addLayout(self.verticalLayout_3)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(parent=ContestStatistics)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.fromDateEdit = QtWidgets.QDateEdit(parent=ContestStatistics)
        self.fromDateEdit.setCalendarPopup(True)
        self.fromDateEdit.setObjectName("fromDateEdit")
        self.verticalLayout.addWidget(self.fromDateEdit)
        self.horizontalLayout.addLayout(self.verticalLayout)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label_2 = QtWidgets.QLabel(parent=ContestStatistics)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_2.addWidget(self.label_2)
        self.toDateEdit = QtWidgets.QDateEdit(parent=ContestStatistics)
        self.toDateEdit.setEnabled(False)
        self.toDateEdit.setCalendarPopup(True)
        self.toDateEdit.setObjectName("toDateEdit")
        self.verticalLayout_2.addWidget(self.toDateEdit)
        self.horizontalLayout.addLayout(self.verticalLayout_2)
        self.verticalLayout_5 = QtWidgets.QVBoxLayout()
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.specificLabel = QtWidgets.QLabel(parent=ContestStatistics)
        self.specificLabel.setEnabled(False)
        self.specificLabel.setObjectName("specificLabel")
        self.verticalLayout_5.addWidget(self.specificLabel)
        self.specificLineEdit = QtWidgets.QLineEdit(parent=ContestStatistics)
        self.specificLineEdit.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.specificLineEdit.sizePolicy().hasHeightForWidth())
        self.specificLineEdit.setSizePolicy(sizePolicy)
        self.specificLineEdit.setMinimumSize(QtCore.QSize(75, 0))
        self.specificLineEdit.setObjectName("specificLineEdit")
        self.verticalLayout_5.addWidget(self.specificLineEdit)
        self.horizontalLayout.addLayout(self.verticalLayout_5)
        self.verticalLayout_4.addLayout(self.horizontalLayout)
        self.tableView = QtWidgets.QTableView(parent=ContestStatistics)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tableView.sizePolicy().hasHeightForWidth())
        self.tableView.setSizePolicy(sizePolicy)
        self.tableView.setMinimumSize(QtCore.QSize(300, 100))
        self.tableView.setEditTriggers(QtWidgets.QAbstractItemView.EditTrigger.NoEditTriggers)
        self.tableView.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.NoSelection)
        self.tableView.setCornerButtonEnabled(False)
        self.tableView.setObjectName("tableView")
        self.tableView.horizontalHeader().setHighlightSections(False)
        self.tableView.verticalHeader().setHighlightSections(False)
        self.verticalLayout_4.addWidget(self.tableView)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.multisLabel = QtWidgets.QLabel(parent=ContestStatistics)
        self.multisLabel.setText("Multis")
        self.multisLabel.setObjectName("multisLabel")
        self.gridLayout.addWidget(self.multisLabel, 0, 0, 1, 1)
        self.multisPlainTextEdit = QtWidgets.QPlainTextEdit(parent=ContestStatistics)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.multisPlainTextEdit.sizePolicy().hasHeightForWidth())
        self.multisPlainTextEdit.setSizePolicy(sizePolicy)
        self.multisPlainTextEdit.setMinimumSize(QtCore.QSize(0, 25))
        self.multisPlainTextEdit.setMaximumSize(QtCore.QSize(16777215, 60))
        self.multisPlainTextEdit.setUndoRedoEnabled(False)
        self.multisPlainTextEdit.setLineWrapMode(QtWidgets.QPlainTextEdit.LineWrapMode.WidgetWidth)
        self.multisPlainTextEdit.setReadOnly(True)
        self.multisPlainTextEdit.setObjectName("multisPlainTextEdit")
        self.gridLayout.addWidget(self.multisPlainTextEdit, 0, 1, 1, 1)
        self.multis2Label = QtWidgets.QLabel(parent=ContestStatistics)
        self.multis2Label.setText("Multis 2")
        self.multis2Label.setObjectName("multis2Label")
        self.gridLayout.addWidget(self.multis2Label, 1, 0, 1, 1)
        self.multis2PlainTextEdit = QtWidgets.QPlainTextEdit(parent=ContestStatistics)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.multis2PlainTextEdit.sizePolicy().hasHeightForWidth())
        self.multis2PlainTextEdit.setSizePolicy(sizePolicy)
        self.multis2PlainTextEdit.setMinimumSize(QtCore.QSize(0, 25))
        self.multis2PlainTextEdit.setMaximumSize(QtCore.QSize(16777215, 60))
        self.multis2PlainTextEdit.setUndoRedoEnabled(False)
        self.multis2PlainTextEdit.setReadOnly(True)
        self.multis2PlainTextEdit.setObjectName("multis2PlainTextEdit")
        self.gridLayout.addWidget(self.multis2PlainTextEdit, 1, 1, 1, 1)
        self.verticalLayout_4.addLayout(self.gridLayout)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_6 = QtWidgets.QLabel(parent=ContestStatistics)
        self.label_6.setObjectName("label_6")
        self.horizontalLayout_2.addWidget(self.label_6)
        self.infosLabel = QtWidgets.QLabel(parent=ContestStatistics)
        self.infosLabel.setObjectName("infosLabel")
        self.horizontalLayout_2.addWidget(self.infosLabel)
        self.line = QtWidgets.QFrame(parent=ContestStatistics)
        self.line.setFrameShape(QtWidgets.QFrame.Shape.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line.setObjectName("line")
        self.horizontalLayout_2.addWidget(self.line)
        self.label_9 = QtWidgets.QLabel(parent=ContestStatistics)
        self.label_9.setObjectName("label_9")
        self.horizontalLayout_2.addWidget(self.label_9)
        self.warningsLabel = QtWidgets.QLabel(parent=ContestStatistics)
        self.warningsLabel.setObjectName("warningsLabel")
        self.horizontalLayout_2.addWidget(self.warningsLabel)
        self.line_2 = QtWidgets.QFrame(parent=ContestStatistics)
        self.line_2.setFrameShape(QtWidgets.QFrame.Shape.VLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line_2.setObjectName("line_2")
        self.horizontalLayout_2.addWidget(self.line_2)
        self.label_7 = QtWidgets.QLabel(parent=ContestStatistics)
        self.label_7.setObjectName("label_7")
        self.horizontalLayout_2.addWidget(self.label_7)
        self.errorsLabel = QtWidgets.QLabel(parent=ContestStatistics)
        self.errorsLabel.setObjectName("errorsLabel")
        self.horizontalLayout_2.addWidget(self.errorsLabel)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.verticalLayout_4.addLayout(self.horizontalLayout_2)
        self.verticalLayout_4.setStretch(1, 3)
        self.verticalLayout_4.setStretch(2, 1)

        self.retranslateUi(ContestStatistics)
        self.contestComboBox.currentTextChanged['QString'].connect(ContestStatistics.contestChanged) # type: ignore
        self.fromDateEdit.dateChanged['QDate'].connect(ContestStatistics.fromDateChanged) # type: ignore
        self.toDateEdit.dateChanged['QDate'].connect(ContestStatistics.toDateChanged) # type: ignore
        self.specificLineEdit.editingFinished.connect(ContestStatistics.specificChanged) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(ContestStatistics)

    def retranslateUi(self, ContestStatistics):
        _translate = QtCore.QCoreApplication.translate
        ContestStatistics.setWindowTitle(_translate("ContestStatistics", "Contest Statistics"))
        self.label_3.setText(_translate("ContestStatistics", "Contest"))
        self.label.setText(_translate("ContestStatistics", "Date from"))
        self.fromDateEdit.setDisplayFormat(_translate("ContestStatistics", "yyyy-MM-dd"))
        self.label_2.setText(_translate("ContestStatistics", "Date to"))
        self.toDateEdit.setDisplayFormat(_translate("ContestStatistics", "yyyy-MM-dd"))
        self.specificLabel.setText(_translate("ContestStatistics", "Specific"))
        self.label_6.setText(_translate("ContestStatistics", "Infos:"))
        self.infosLabel.setText(_translate("ContestStatistics", "0"))
        self.label_9.setText(_translate("ContestStatistics", "Warnings:"))
        self.warningsLabel.setText(_translate("ContestStatistics", "0"))
        self.label_7.setText(_translate("ContestStatistics", "Errors:"))
        self.errorsLabel.setText(_translate("ContestStatistics", "0"))



# DragonLog (c) 2023 by Andreas Schawo is licensed under CC BY-SA 4.0.
# To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/4.0/

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum

import requests
import xmltodict
from adif_file import adi

from .Logger import Logger


class CallBookType(Enum):
    HamQTH = 'HamQTH.com'
    QRZCQ = 'QRZCQ.com'
    QRZ = 'QRZ.com'


@dataclass
class CallBookData:
    callsign: str
    nickname: str
    locator: str
    qth: str
    qsl_via: str
    qsl_bureau: bool
    qsl_direct: bool
    qsl_eqsl: bool
    qsl_lotw: bool
    darc_dok: str


class CommunicationException(Exception):
    pass


class RequestException(Exception):
    pass


class LoginException(Exception):
    pass


class SessionExpiredException(Exception):
    pass


class QSORejectedException(Exception):
    pass


class MissingADIFFieldException(Exception):
    pass


class CallsignNotFoundException(Exception):
    pass


class AbstractCallBook(ABC):
    """Standardised interface to access diffrent callbook services

    :param logger: the application log
    :param prog_name: the calling programs name"""

    def __init__(self, logger: Logger, prog_name: str):
        self.__program_str__ = prog_name
        self.__session__: str = ''

        self.log = logging.getLogger(type(self).__name__)
        self.log.setLevel(logger.loglevel)
        self.log.addHandler(logger)
        self.log.debug('Initialising...')

    @property
    @abstractmethod
    def required_fields(self) -> tuple:
        """A list of required fields to create a logbook entry"""
        pass

    @property
    @abstractmethod
    def __url__(self) -> str:
        pass

    def _get_(self, params: dict) -> dict:
        r = requests.get(self.__url__, params=params)

        if r.status_code == 200:
            return xmltodict.parse(r.text)
        else:
            raise CommunicationException(f'{self.callbook_type.name} error: HTTP-Error {r.status_code}')

    @property
    @abstractmethod
    def callbook_type(self) -> CallBookType:
        """The type of the callbook"""
        pass

    @abstractmethod
    def __login__(self, username: str, password: str) -> str:
        pass

    def login(self, username: str, password: str):
        """Login and retreive a session key
        :param username: the username used with this service
        :param password: the password used with this service"""
        if not username or not password:
            raise LoginException('Username or password missing')

        self.__session__ = self.__login__(username, password)

    @property
    def is_loggedin(self) -> bool:
        """Check if already logged in"""
        return bool(self.__session__)

    @abstractmethod
    def __get_dataset__(self, callsign: str):
        pass

    def get_dataset(self, callsign: str) -> CallBookData:
        """Perform a data search for a callsign search
        :param callsign: the callsign to be searched for
        :return: the callbook data set"""
        try:
            return self.__get_dataset__(callsign)
        except SessionExpiredException as exc:
            self.__session__ = ''
            raise exc

    @property
    @abstractmethod
    def has_logbook(self) -> bool:
        """Is a logbook available for this service"""
        pass

    @abstractmethod
    def __upload_log__(self, username: str, password: str, adif: str):
        pass

    def upload_log(self, username: str, password: str, adif_data: dict):
        """Upload logbook data
        :param username: the username used with this service
        :param password: the password used with this service
        :param adif_data: the data to be converted to ADI and uploaded"""

        if not username or not password:
            raise LoginException('Username or password missing')

        for field in self.required_fields:
            if field not in adif_data['RECORDS'][0]:
                raise MissingADIFFieldException(field)

        adif_data = adif_data.copy()
        adif_data['RECORDS'] = [adif_data['RECORDS'][0]]

        self.__upload_log__(username, password, adi.dumps(adif_data, 'ADIF Export by DragonLog'))


class HamQTHCallBook(AbstractCallBook):
    def __init__(self, logger: Logger, prog_name: str):
        super().__init__(logger, prog_name)

    @property
    def required_fields(self) -> tuple:
        return 'QSO_DATE', 'TIME_ON', 'CALL', 'MODE', 'BAND', 'RST_SENT', 'RST_RCVD'

    @property
    def __url__(self):
        return 'https://www.hamqth.com/xml.php'

    @property
    def callbook_type(self) -> CallBookType:
        return CallBookType.HamQTH

    def __login__(self, username: str, password: str) -> str:
        if not username or not password:
            raise LoginException('Username or password missing')

        try:
            res = self._get_({'u': username, 'p': password})
        except CommunicationException as exc:
            raise LoginException(str(exc))

        match res:
            case {'HamQTH': {'session': {'error': error}}}:
                raise LoginException(f"HamQTH error: {error}")
            case {'HamQTH': {'session': {'session_id': session_id}}}:
                return session_id
            case _:
                raise LoginException(f'HamQTH error: Unknown data format {res}')

    def __get_dataset__(self, callsign: str) -> CallBookData | None:
        try:
            self.log.debug(f'Searching {callsign}...')
            res = self._get_({'id': self.__session__,
                              'prg': self.__program_str__,
                              'callsign': callsign})
        except CommunicationException as exc:
            raise RequestException(str(exc))

        match res:
            case {'HamQTH': {'session': {'error': error}}}:
                if error == 'Session does not exist or expired':
                    raise SessionExpiredException('HamQTH')
                elif error == 'Callsign not found':
                    raise CallsignNotFoundException(callsign)
                else:
                    raise RequestException(f"HamQTH error: {error}")
            case {'HamQTH': {'search': data}}:
                if data:
                    return CallBookData(
                        callsign,
                        data.get('nick', ''),
                        data.get('grid', ''),
                        data.get('qth', ''),
                        data.get('qsl_via', ''),
                        data.get('qsl', '') == 'Y',
                        data.get('qsldirect', '') == 'Y',
                        data.get('eqsl', '') == 'Y',
                        data.get('lotw', '') == 'Y',
                        data.get('dok', ''),
                    )
                else:
                    return None
            case _:
                raise RequestException(f'HamQTH error: Unknown data format {res}')

    @property
    def has_logbook(self) -> bool:
        return True

    def __upload_log__(self, username: str, password: str, adif: str):
        if not username or not password:
            raise LoginException('Username or password missing')

        data = {
            'u': username,
            'p': password,
            'adif': adif,
            'prg': self.__program_str__,
            'cmd': 'insert'
        }

        r = requests.post('https://www.hamqth.com/qso_realtime.php', data=data)

        if r.status_code == 200:
            return
        elif r.status_code == 400:
            raise QSORejectedException(r.text)
        elif r.status_code == 403:
            raise LoginException(r.text)
        else:
            self.log.debug(f'HamQTH communication error #{r.status_code} "{r.text}"')
            raise CommunicationException(f'HamQTH error: HTTP-Error {r.status_code}')


class QRZCQCallBook(AbstractCallBook):
    def __init__(self, logger: Logger, prog_name: str):
        super().__init__(logger, prog_name)

    @property
    def required_fields(self) -> tuple:
        raise NotImplementedError('QRZCQ.com API for uploading is currently not stable')

    @property
    def __url__(self):
        return 'https://ssl.qrzcq.com/xml'

    @property
    def callbook_type(self) -> CallBookType:
        return CallBookType.QRZCQ

    def __login__(self, username: str, password: str) -> str:
        if not username or not password:
            raise LoginException('Username or password missing')

        try:
            res = self._get_({'username': username,
                              'password': password,
                              'agent': self.__program_str__})
        except CommunicationException as exc:
            raise LoginException(str(exc))

        match res:
            case {'QRZCQDatabase': {'Session': {'Error': error}}}:
                raise LoginException(f"QRZCQ error: {error}")
            case {'QRZCQDatabase': {'Session': {'Key': session_id}}}:
                return session_id
            case _:
                raise LoginException(f'QRZCQ error: Unknown data format {res}')

    def __get_dataset__(self, callsign: str) -> CallBookData | None:
        try:
            self.log.debug(f'Searching {callsign}...')
            res = self._get_({'s': self.__session__,
                              'callsign': callsign,
                              'agent': self.__program_str__})
        except CommunicationException as exc:
            raise RequestException(str(exc))

        match res:
            case {'QRZCQDatabase': {'Session': {'Error': error}}}:
                if error == 'Session does not exist or expired':
                    raise SessionExpiredException('QRZCQ')
                elif error.startswith('Not found') or error.startswith('Callsign Empty'):
                    raise CallsignNotFoundException(callsign)
                else:
                    raise RequestException(f"QRZCQ error: {error}")
            case {'QRZCQDatabase': {'Callsign': data}}:
                if data:
                    return CallBookData(
                        callsign,
                        data.get('name', ''),
                        data.get('locator', ''),
                        data.get('qth', ''),
                        data.get('manager', ''),
                        data.get('bqsl', '') == '1',
                        data.get('mqsl', '') == '1',
                        data.get('eqsl', '') == '1',
                        data.get('lotw', '') == '1',
                        data.get('dok', ''),
                    )
                else:
                    return None
            case _:
                raise RequestException(f'QRZCQ error: Unknown data format {res}')

    @property
    def has_logbook(self) -> bool:
        return False

    def __upload_log__(self, username: str, password: str, adif: str):
        raise NotImplementedError('QRZCQ.com API for uploading is currently not stable')


class QRZCallBook(AbstractCallBook):
    def __init__(self, logger: Logger, prog_name: str):
        super().__init__(logger, prog_name)

    @property
    def required_fields(self) -> tuple:
        raise NotImplementedError('QRZ.com API for uploading is not available')

    @property
    def __url__(self):
        return 'https://xmldata.qrz.com/xml/1.34/'

    @property
    def callbook_type(self) -> CallBookType:
        return CallBookType.QRZ

    def __login__(self, username: str, password: str) -> str:
        if not username or not password:
            raise LoginException('Username or password missing')

        try:
            res = self._get_({'username': username,
                              'password': password,
                              'agent': self.__program_str__})
        except CommunicationException as exc:
            raise LoginException(str(exc))

        match res:
            case {'QRZDatabase': {'Session': {'Error': error}}}:
                raise LoginException(f"QRZ error: {error}")
            case {'QRZDatabase': {'Session': {'Key': session_id}}}:
                return session_id
            case _:
                raise LoginException(f'QRZ error: Unknown data format {res}')

    def __get_dataset__(self, callsign: str) -> CallBookData | None:
        try:
            self.log.debug(f'Searching {callsign}...')
            res = self._get_({'s': self.__session__,
                              'callsign': callsign,
                              # 'agent': self.__program_str__
                              })
        except CommunicationException as exc:
            raise RequestException(str(exc))

        match res:
            case {'QRZDatabase': {'Session': {'Error': error}}}:
                if error == 'Session does not exist or expired':
                    raise SessionExpiredException('QRZ')
                elif error.startswith('Not found') or error.startswith('Callsign Empty'):
                    raise CallsignNotFoundException(callsign)
                else:
                    raise RequestException(f"QRZ error: {error}")
            case {'QRZDatabase': {'Callsign': data}}:
                if data:
                    return CallBookData(
                        callsign,
                        data.get('name_fmt', data.get('fname', '')),
                        data.get('grid', ''),
                        data.get('addr2', ''),
                        data.get('qslmgr', ''),
                        False,  # not available
                        data.get('mqsl', '') == '1',
                        data.get('eqsl', '') == '1',
                        data.get('lotw', '') == '1',
                        '',  # not available
                    )
                else:
                    return None
            case _:
                raise RequestException(f'QRZ error: Unknown data format {res}')

    @property
    def has_logbook(self) -> bool:
        return False

    def __upload_log__(self, username: str, password: str, adif: str):
        raise NotImplementedError('QRZ.com API for uploading is not available')



# DragonLog (c) 2025 by Andreas Schawo is licensed under CC BY-SA 4.0.
# To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/4.0/
"""Contains main DARC contests"""

from .base import (ContestLogCBR, ContestLogEDI, CBRRecord, Address, BandStatistics, BAND_FROM_CBR, ExchangeData,
                   CategoryMode, CategoryBand, CategoryPower, CategoryOperator, CategoryAssisted, CategoryTransmitter)
from dragonlog.cty import CountryData


class DARCUKWFruehlingsContest(ContestLogEDI):
    contest_name = 'DARC UKW Frühlingswettbewerb'
    contest_year = '2025'
    contest_update = '2025-04-07'

    @classmethod
    def valid_modes(cls) -> tuple[CategoryMode, ...]:
        return CategoryMode.SSB, CategoryMode.CW, CategoryMode.FM

    @classmethod
    def valid_bands(cls) -> tuple[CategoryBand, ...]:
        return (CategoryBand.B_2M, CategoryBand.B_70CM, CategoryBand.B_23CM, CategoryBand.B_13CM,
                CategoryBand.B_9CM, CategoryBand.B_6CM, CategoryBand.B_3CM, CategoryBand.B_1_25CM,
                CategoryBand.B_6MM, CategoryBand.B_4MM, CategoryBand.B_2_5MM, CategoryBand.B_2MM, CategoryBand.B_1MM)

    @classmethod
    def descr_specific(cls) -> str:
        return 'DOK'

    @classmethod
    def valid_operator(cls) -> tuple[CategoryOperator, ...]:
        return CategoryOperator.SINGLE_OP, CategoryOperator.MULTI_OP, CategoryOperator.TRAINEE, CategoryOperator.CHECKLOG


class DARCUKWSommerFDContest(ContestLogEDI):
    contest_name = 'DARC UKW-Sommer-Fieldday'
    contest_year = '2025'
    contest_update = '2025-08-02'

    @classmethod
    def valid_modes(cls) -> tuple[CategoryMode, ...]:
        return CategoryMode.SSB, CategoryMode.CW, CategoryMode.FM

    @classmethod
    def valid_bands(cls) -> tuple[CategoryBand, ...]:
        return (CategoryBand.B_2M, CategoryBand.B_70CM,
                CategoryBand.B_23CM, CategoryBand.B_13CM, CategoryBand.B_9CM, CategoryBand.B_6CM)

    @classmethod
    def valid_operator(cls) -> tuple[CategoryOperator, ...]:
        return CategoryOperator.SINGLE_OP, CategoryOperator.TRAINEE, CategoryOperator.CHECKLOG


class DARCUKWContest(ContestLogEDI):
    contest_name = 'DARC UKW-Wettbewerb'
    contest_year = '2025'
    contest_update = '2025-04-07'

    @classmethod
    def valid_modes(cls) -> tuple[CategoryMode, ...]:
        return CategoryMode.SSB, CategoryMode.CW, CategoryMode.FM

    @classmethod
    def valid_bands(cls) -> tuple[CategoryBand, ...]:
        return (CategoryBand.B_2M, CategoryBand.B_70CM, CategoryBand.B_23CM, CategoryBand.B_13CM,
                CategoryBand.B_9CM, CategoryBand.B_6CM, CategoryBand.B_3CM, CategoryBand.B_1_25CM,
                CategoryBand.B_6MM, CategoryBand.B_4MM, CategoryBand.B_2_5MM, CategoryBand.B_2MM, CategoryBand.B_1MM)

    @classmethod
    def descr_specific(cls) -> str:
        return 'DOK'

    @classmethod
    def is_single_day(cls) -> bool:
        return False

    @classmethod
    def valid_power(cls) -> tuple[CategoryPower, ...]:
        return CategoryPower.NONE, CategoryPower.LOW

    @classmethod
    def valid_operator(cls) -> tuple[CategoryOperator, ...]:
        return CategoryOperator.SINGLE_OP, CategoryOperator.MULTI_OP, CategoryOperator.TRAINEE, CategoryOperator.CHECKLOG


class DARCOsterContest(ContestLogCBR):
    contest_name = 'DARC-Ostercontest'
    contest_year = '2025'
    contest_update = '2025-04-13'
    contest_exch_fmt = 'DOK'

    def __init__(self, callsign: str, name: str, club: str, address: Address, email: str, locator: str,
                 band: type[CategoryBand], mode: type[CategoryMode],
                 pwr: type[CategoryPower] = CategoryPower.HIGH,
                 cat_operator: type[CategoryOperator] = CategoryOperator.SINGLE_OP,
                 assisted: type[CategoryAssisted] = CategoryAssisted.NON_ASSISTED,
                 tx: type[CategoryTransmitter] = CategoryTransmitter.ONE,
                 operators: list[str] = None, specific: str = '', skip_id: bool = False,
                 skip_warn: bool = False, logger=None, **params):
        super().__init__(callsign, name, club, address, email, locator,
                         band, mode, pwr, cat_operator,
                         assisted, tx, operators, specific, skip_id, skip_warn, logger)

        self.cty: CountryData | None = params.get('cty', None)
        if not self.cty or type(self.cty) is not CountryData:
            self.cty = None
            self.log.error('Error with cty data')

        self.__band_multis__: dict[str, set[str]] = {}

    def build_record(self, adif_rec) -> CBRRecord:
        adif_rec['STX_STRING'] = self.__header__['SPECIFIC'].upper()
        rec = super().build_record(adif_rec)
        return rec

    def process_points(self, rec: CBRRecord):
        try:
            qso_point = 1

            self.__rated__ += 1
            band = BAND_FROM_CBR[rec.band]
            if band not in self.__band_multis__:
                self.__band_multis__[band] = set()

            if rec.rcvd_exch.upper() != 'NM':
                multi = 'DOK_' + rec.mode + rec.rcvd_exch.upper()
                if multi not in self.__band_multis__[band]:
                    self.__multis__.add(multi)
                self.__band_multis__[band].add(multi)
            else:
                self.info(f'DOK not counted as multi "{rec.rcvd_exch.upper()}"')

            if self.cty:
                pfx = self.cty.prefix(rec.call)
                if pfx:
                    multi = 'PFX_' + rec.mode + pfx
                    if multi not in self.__band_multis__[band]:
                        self.__multis__.add(multi)
                    self.__band_multis__[band].add(multi)
            else:
                self.log.warning('Could not process prefix due to missing cty data')

            self.__points__ += qso_point

            if band in self.__stats__:
                self.__stats__[band].qsos += 1
                self.__stats__[band].rated += 1
                self.__stats__[band].points += qso_point
                self.__stats__[band].multis = len(self.__band_multis__[band])
            else:
                self.__stats__[band] = BandStatistics(1, 1, 1, len(self.__band_multis__[band]), 0, qso_point)

            self.__header__['CLAIMED-SCORE'] = str(self.claimed_points)
        except Exception:
            self.exception()

    @property
    def statistics(self) -> dict[str, BandStatistics]:
        qsos = 0
        rated = 0
        points = 0
        multis = 0
        claimed = 0
        for b in self.__stats__:
            self.__stats__[b].summary = self.__stats__[b].points * self.__stats__[b].multis
            qsos += self.__stats__[b].qsos
            rated += self.__stats__[b].rated
            points += self.__stats__[b].points
            multis += self.__stats__[b].multis
            claimed += self.__stats__[b].summary

        self.__stats__['Total'] = BandStatistics(qsos, rated, points, multis, 0, claimed)

        return self.__stats__

    @property
    def file_name(self) -> str:
        return f'{self.__contest_date__}_{self.__header__["CALLSIGN"]}-{self.__header__["SPECIFIC"]}-{self.__header__["CATEGORY-BAND"]}.cbr'

    @classmethod
    def valid_modes(cls) -> tuple[CategoryMode, ...]:
        return CategoryMode.SSB, CategoryMode.MIXED, CategoryMode.CW

    @classmethod
    def valid_bands(cls) -> tuple[CategoryBand, ...]:
        return CategoryBand.B_ALL, CategoryBand.B_80M, CategoryBand.B_40M

    @classmethod
    def descr_specific(cls) -> str:
        return 'DOK'

    @classmethod
    def needs_specific(cls) -> bool:
        return True

    @classmethod
    def valid_power(cls) -> tuple[CategoryPower, ...]:
        return CategoryPower.HIGH, CategoryPower.LOW

    @classmethod
    def valid_operator(cls) -> tuple[CategoryOperator, ...]:
        return CategoryOperator.SINGLE_OP, CategoryOperator.CHECKLOG

    @staticmethod
    def extract_exchange(exchange: str) -> ExchangeData | None:
        if exchange:
            return ExchangeData(darc_dok=exchange.strip())
        else:
            return None

    @staticmethod
    def prepare_exchange(exchange: ExchangeData):
        return f'{exchange.darc_dok}'
